#include "geometrylib.h"

#include "mathlib.h"
#include "arrays.h"

#include <stdio.h>
#include <stdlib.h>
#include <math.h>


namespace gelibspace {

///     ZDROJE
/// http://softsurfer.com/algorithm_archive.htm
///

///**
//   Function computes edge points of common perpendicular of two line segments l1 and l2.
//   The line segment l1 goes from point l1a to l1b.
//   l2 (l2a l2b)
//   
//   Return 0 - both edge points of common perpendicular is lying  inside of the line segments l1 and l2
//   Return 1 -      edge point  of common perpendicular is lying outside of the line segment l1
//   Return 2 -      edge point  of common perpendicular is lying outside of the line segment l2
//   Return 3 - both edge points of common perpendicular is lying outside of the line segments l1 and l2
//*/
//int give_edge_points_commperpend_skewlines (const double *l1a, const double *l1b, const double *l2a, const double *l2b, double *p1,  double *p2)
//{
//  Dvctr x(3), y(3), z(3);
//  
//  x.be_point2point( l1a, l1b );
//  y.be_point2point( l2a, l2b );
//  z.be_vectproduct( x, y );
//  
//  double m[9], r[3], l[3];
//  
//  m[0] = -x(0);  m[1] = y(0);  m[2] = z(0);   r[0] = l1a[0] - l2a[0];
//  m[3] = -x(1);  m[4] = y(1);  m[5] = z(1);   r[1] = l1a[1] - l2a[1];
//  m[6] = -x(2);  m[7] = y(2);  m[8] = z(2);   r[2] = l1a[2] - l2a[2];
//  
//  solve_3 (m, r, l);
//  
//  p1[0] = l1a[0] + l[0] * x(0);    p2[0] = l2a[0] + l[1] * y(0);
//  p1[1] = l1a[1] + l[0] * x(1);    p2[1] = l2a[1] + l[1] * y(1);
//  p1[2] = l1a[2] + l[0] * x(2);    p2[2] = l2a[2] + l[1] * y(2);
//  
//  int answer = 0;
//  if ( l[0] < 0.0  ||  1.0 < l[0] )  answer += 1;
//  if ( l[1] < 0.0  ||  1.0 < l[1] )  answer += 2;
//  
//  return answer;
//}


/**
   Function computes natural and cartesian coordinates of points of intersection of line and warped rectangle
   answer: -1 = infinite number of points                                               
           0,1,2 = number of points							           V              
   											 	    \             
   @param zero - accuracy of solving                                                              D--\---.------C 
   @param norm - norm is on order equal to diagonal of rectangle                                 /    \ .      /  
   @param A,B,C,D - arrays of cartesian coordinates of rectangle vertexes, dimension is 3   eta /......\..... /   
   @param U,V - arrays of cartesian coordinates of vertexes of line, dimension is 3            /      . +    /    
   @param xi,eta - arrays of natural coordinates of "points", dimension is 2                  /      .   +  /     
   @param t - array of position of intersection points between U(t=-1) and V(t=1)            /      .     +/      
   @param I1,I2 - arrays of cartesian coordinates of "points", dimension is 3               /      .      /\      
   											   A-------------B  \     
   created  4.4.2003, Ladislav Svoboda, termit@cml.fsv.cvut.cz                                  ksi          U    
*/
long intersec_rectangle3d_line (double zero, double norm,
                                const PoinT *A, const PoinT *B, const PoinT *C, const PoinT *D,
                                const PoinT *U, const PoinT *V,
                                double *ksi, double *eta, double *t,
				PoinT *I1, PoinT *I2)
{
  long i,ii,jj,kk,nr;
  
  VectoR a,b,c,d,e;
  double aa[2],bb[2],cc[2],dd[2];
  
  if (!norm) {
    a.beP2P(C,A);
    b.beP2P(D,B);
    norm = 0.5 * ( sqrt(a.x*a.x+a.y*a.y+a.z*a.z) + sqrt(b.x*b.x+b.y*b.y+b.z*b.z) );
  }
  
  a.beP2P(B,A);  a.add(b.beP2P(D,C));
  b.beP2P(A,B);
  c.beP2P(A,D);
  d.beP2P(V,U);
  e.beP2P(U,A);
  
  if (fabs(d.x) > fabs(d.y)) if (fabs(d.x) > fabs(d.z)) { ii=1; jj=2; kk=0; }
                             else                       { ii=0; jj=1; kk=2; }
  else                       if (fabs(d.y) > fabs(d.z)) { ii=2; jj=0; kk=1; }
                             else                       { ii=0; jj=1; kk=2; }
  
  if (fabs(d[kk]) < norm*zero) _errorr("Length of line is 0.0");
  
  aa[0] = a[ii]*d[kk] - a[kk]*d[ii];  aa[1] = a[jj]*d[kk] - a[kk]*d[jj];
  bb[0] = b[ii]*d[kk] - b[kk]*d[ii];  bb[1] = b[jj]*d[kk] - b[kk]*d[jj];
  cc[0] = c[ii]*d[kk] - c[kk]*d[ii];  cc[1] = c[jj]*d[kk] - c[kk]*d[jj];
  dd[0] = e[ii]*d[kk] - e[kk]*d[ii];  dd[1] = e[jj]*d[kk] - e[kk]*d[jj];
  
  nr = solv_2nle (norm*norm*zero*zero,aa,bb,cc,dd,ksi,eta);
  
  // nefunguje solv_2nle resic
  if (nr == -1)  errol;
  if (nr == -2) {
    if (0 < intersect_RayTriangle (zero, U, V, A, B, C, I1)) errol;
    if (0 < intersect_RayTriangle (zero, U, V, A, C, D, I1)) errol;
    if (0 < intersect_RayTriangle (zero, U, V, A, B, D, I1)) errol;
    if (0 < intersect_RayTriangle (zero, U, V, B, C, D, I1)) errol;
    nr = 0;
  }
  
  for (i=0; i<nr; i++) {
    t[i] = (a[kk]*ksi[i]*eta[i] + b[kk]*ksi[i] + c[kk]*eta[i] + e[kk]) / -d[kk] ;
    t[i]   =  2.0*(  t[i]-0.5);
    ksi[i] = -2.0*(ksi[i]-0.5);
    eta[i] = -2.0*(eta[i]-0.5);
  }
  
  if (nr > 0)   I1->bePointAtAbscissa (U,V,t[0]);
  if (nr > 1)   I2->bePointAtAbscissa (U,V,t[1]);
  
  if (0){
    fprintf (stdout,"\n number of results %ld\n",nr);
    if (nr > 0)  fprintf (stdout,"\n  ksi =  %5.2f  eta =  %5.2f  t = %5.2f  x = %15.7f  y = %15.7f  z = %15.7f\n", ksi[0],eta[0],t[0],I1->x,I1->y,I1->z);
    if (nr > 1)  fprintf (stdout,"\n  ksi =  %5.2f  eta =  %5.2f  t = %5.2f  x = %15.7f  y = %15.7f  z = %15.7f\n", ksi[1],eta[1],t[1],I2->x,I2->y,I2->z);
  }
  
  return nr;
}


/**
   Function computes natural coordinates of 'point' on 'element',
   'point' is entered in cartesian coordinates, 'element' is 3d brick=cube=hexahedron.
   
   @param zero   - tolerance of natural coordinates on boundarys,(approximately 1.0e-5)
   @param elem   - pointer on 'element'
   @param point  - array of coordinates of 'point',dimension is 3
   @param answer - array of dimension 3
   
   created  xx.x.xxxx, Dan Rypl, dr@cml.fsv.cvut.cz
*/
long nc_brick_3d (double zero, double x[], double y[], double z[], const PoinT *point, PoinT *answer)
{
  long i,nite;
  double a1,a2,a3,a4,a5,a6,a7,a8,b1,b2,b3,b4,b5,b6,b7,b8,c1,c2,c3,c4,c5,c6,c7,c8;
  double xp,yp,zp,u,v,w;
  double r[3],p[9],delta[3];
  
  xp = point->x;                                                         //                                                       ksi
  yp = point->y;                                                         //                                                     7
  zp = point->z;                                                         //                                                    /
                                                                         //                                                   /
  a1 =  x[0] + x[1] + x[2] + x[3] + x[4] + x[5] + x[6] + x[7];           //                        8----------------7----------------7
  a2 = -x[0] - x[1] + x[2] + x[3] - x[4] - x[5] + x[6] + x[7];	         //                       /|                        *       /|
  a3 = -x[0] + x[1] + x[2] - x[3] - x[4] + x[5] + x[6] - x[7];	         //                      / |                       *       / |
  a4 =  x[0] + x[1] + x[2] + x[3] - x[4] - x[5] - x[6] - x[7];	         //                     /  |                      *       /  |
  a5 =  x[0] - x[1] + x[2] - x[3] + x[4] - x[5] + x[6] - x[7];	         //                    /   |                     *       /   |
  a6 = -x[0] - x[1] + x[2] + x[3] + x[4] + x[5] - x[6] - x[7];	         //                   /    |                    *       /    |
  a7 = -x[0] + x[1] + x[2] - x[3] + x[4] - x[5] - x[6] + x[7];	         //                  8     |          2        *       6     |
  a8 =  x[0] - x[1] + x[2] - x[3] - x[4] + x[5] - x[6] + x[7];	         //                 /      |                  *       /      |
                                                                         //                /       |                 *       /       |
  b1 =  y[0] + y[1] + y[2] + y[3] + y[4] + y[5] + y[6] + y[7];	         //               /       12                5       /       11
  b2 = -y[0] - y[1] + y[2] + y[3] - y[4] - y[5] + y[6] + y[7];	         //              /         |               *       /         |
  b3 = -y[0] + y[1] + y[2] - y[3] - y[4] + y[5] + y[6] - y[7];	         //             /          |              *       /          |
  b4 =  y[0] + y[1] + y[2] + y[3] - y[4] - y[5] - y[6] - y[7];	         //            5----------------5----------------6           |
  b5 =  y[0] - y[1] + y[2] - y[3] + y[4] - y[5] + y[6] - y[7];	         //            |           |            *        |           |
  b6 = -y[0] - y[1] + y[2] + y[3] + y[4] + y[5] - y[6] - y[7];	         //            |           |           *         |           |    eta
  b7 = -y[0] + y[1] + y[2] - y[3] + y[4] - y[5] - y[6] + y[7];	         //            |     6     |          O * * * * *|* * *4----------->
  b8 =  y[0] - y[1] + y[2] - y[3] - y[4] + y[5] - y[6] + y[7];	         //            |           |          *          |           |
                                                                         //            |           |          *          |           |
  c1 =  z[0] + z[1] + z[2] + z[3] + z[4] + z[5] + z[6] + z[7];	         //            |           4----------*-----3----|-----------3
  c2 = -z[0] - z[1] + z[2] + z[3] - z[4] - z[5] + z[6] + z[7];	         //            |          /           *          |          / 
  c3 = -z[0] + z[1] + z[2] - z[3] - z[4] + z[5] + z[6] - z[7];	         //            |         /            *          |         /  
  c4 =  z[0] + z[1] + z[2] + z[3] - z[4] - z[5] - z[6] - z[7];	         //            9        /       3     *         10        /   
  c5 =  z[0] - z[1] + z[2] - z[3] + z[4] - z[5] + z[6] - z[7];	         //            |       /              *          |       /    
  c6 = -z[0] - z[1] + z[2] + z[3] + z[4] + z[5] - z[6] - z[7];	         //            |      /               *          |      /     
  c7 = -z[0] + z[1] + z[2] - z[3] + z[4] - z[5] - z[6] + z[7];	         //            |     4                1          |     2      
  c8 =  z[0] - z[1] + z[2] - z[3] - z[4] + z[5] - z[6] + z[7];           //            |    /                 *          |    /       
                                                                         //            |   /                  *          |   /        
  // setup initial guess					         //            |  /                   *          |  /         
  answer->zero();                        			         //            | /                    *          | /          
                                                                         //            |/                     *          |/           
  // apply Newton-Raphson to solve the problem			         //            1----------------1----------------2            
  nite = 0;							         //                                   |                       
  do {								         //                                   |                       
    if ((++nite) > 10) {					         //                                   |                       
      fprintf (stderr,"nc_hexa_3d: no convergence after 10 iterations"); //                                   V dzeta                 
      return 0;
    }
    u = answer->x;
    v = answer->y;
    w = answer->z;
    
    // compute the residual
    r[0] = a1 + u * a2 + v * a3 + w * a4 + u * v * a5 + u * w * a6 + v * w * a7 + u * v * w * a8 - 8.0 * xp;
    r[1] = b1 + u * b2 + v * b3 + w * b4 + u * v * b5 + u * w * b6 + v * w * b7 + u * v * w * b8 - 8.0 * yp;
    r[2] = c1 + u * c2 + v * c3 + w * c4 + u * v * c5 + u * w * c6 + v * w * c7 + u * v * w * c8 - 8.0 * zp;
    
    // check for convergence
    if ((r[0]*r[0]+r[1]*r[1]+r[2]*r[2]) < 1.e-20) break;      // sqrt(1.e-20) = 1.e-10
    
    p[0] = a2 + v * a5 + w * a6 + v * w * a8;                            //         ***   JE TO ODVOZENO PRO TYTO BAZOVE FCE   ***
    p[1] = a3 + u * a5 + w * a7 + u * w * a8;			         //    N[1] = 0.125 * (1.0 - ksi) * (1.0 - eta) * (1.0 + dzeta);
    p[2] = a4 + u * a6 + v * a7 + u * v * a8;			         //    N[2] = 0.125 * (1.0 - ksi) * (1.0 + eta) * (1.0 + dzeta);
    p[3] = b2 + v * b5 + w * b6 + v * w * b8;			         //    N[3] = 0.125 * (1.0 + ksi) * (1.0 + eta) * (1.0 + dzeta);
    p[4] = b3 + u * b5 + w * b7 + u * w * b8;			         //    N[4] = 0.125 * (1.0 + ksi) * (1.0 - eta) * (1.0 + dzeta);
    p[5] = b4 + u * b6 + v * b7 + u * v * b8;			         //    N[5] = 0.125 * (1.0 - ksi) * (1.0 - eta) * (1.0 - dzeta);
    p[6] = c2 + v * c5 + w * c6 + v * w * c8;			         //    N[6] = 0.125 * (1.0 - ksi) * (1.0 + eta) * (1.0 - dzeta);
    p[7] = c3 + u * c5 + w * c7 + u * w * c8;			         //    N[7] = 0.125 * (1.0 + ksi) * (1.0 + eta) * (1.0 - dzeta);
    p[8] = c4 + u * c6 + v * c7 + u * v * c8;                            //    N[8] = 0.125 * (1.0 + ksi) * (1.0 - eta) * (1.0 - dzeta);
    
    // solve for corrections
    solve_3 (p,r,delta);
    
    // update guess
    answer->x -= delta[0];
    answer->y -= delta[1];
    answer->z -= delta[2];
    
  } while (1);
  
  // test if inside
  for (i=0;i<3;i++)
    if ((answer[0][i] < -(1.0+zero)) || (answer[0][i] > (1.0+zero)))  return 0;
  
  return 1;
}


//  long nc_brick_3d (double zero,const element *elem,const double *point,double *answer)
//  {
//    long i,nite;
//    double x[8],y[8],z[8];
//    double a1,a2,a3,a4,a5,a6,a7,a8,b1,b2,b3,b4,b5,b6,b7,b8,c1,c2,c3,c4,c5,c6,c7,c8;
//    double xp,yp,zp,u,v,w;
//    double r[3],p[9],delta[3];
//    
//    for (i=0;i<8;i++){
//      x[i] = Nodes[elem->give_node(i]]->coord[0];  //                                                           *****
//      y[i] = Nodes[elem->give_node(i]]->coord[1];  //                                                           *(1)*
//      z[i] = Nodes[elem->give_node(i]]->coord[2];  //                                                           *****
//    }
//    
//    xp = point[0];
//    yp = point[1];
//    zp = point[2];
//    
//    a1 = x[0] + x[1] + x[2] + x[3] + x[4]	+ x[5] + x[6] + x[7];            //                        8----------------7----------------7
//    a2 = x[0] - x[1] - x[2] + x[3] + x[4]	- x[5] - x[6] + x[7];	         //                       /|                                /|
//    a3 = x[0] + x[1] - x[2] - x[3] + x[4]	+ x[5] - x[6] - x[7];	         //                      / |                               / |
//    a4 = x[0] + x[1] + x[2] + x[3] - x[4]	- x[5] - x[6] - x[7];	         //                     /  |                              /  |
//    a5 = x[0] - x[1] + x[2] - x[3] + x[4]	- x[5] + x[6] - x[7];	         //                    /   |                             /   |
//    a6 = x[0] - x[1] - x[2] + x[3] - x[4]	+ x[5] + x[6] - x[7];	         //                   /    |                            /    |
//    a7 = x[0] + x[1] - x[2] - x[3] - x[4]	- x[5] + x[6] + x[7];	         //                  8     |          2                6     |
//    a8 = x[0] - x[1] + x[2] - x[3] - x[4]	+ x[5] - x[6] + x[7];	         //                 /      |                          /      |
//                                                                           //                /       |                         /       |
//    b1 = y[0] + y[1] + y[2] + y[3] + y[4]	+ y[5] + y[6] + y[7];	         //               /       12                5       /       11
//    b2 = y[0] - y[1] - y[2] + y[3] + y[4]	- y[5] - y[6] + y[7];	         //              /         |                       /         |
//    b3 = y[0] + y[1] - y[2] - y[3] + y[4]	+ y[5] - y[6] - y[7];	         //             /          |                      /          |
//    b4 = y[0] + y[1] + y[2] + y[3] - y[4]	- y[5] - y[6] - y[7];	         //            5----------------5----------------6           |
//    b5 = y[0] - y[1] + y[2] - y[3] + y[4]	- y[5] + y[6] - y[7];	         //            |           |                     |           |
//    b6 = y[0] - y[1] - y[2] + y[3] - y[4]	+ y[5] + y[6] - y[7];	         // ksi        |           |                     |           |
//    b7 = y[0] + y[1] - y[2] - y[3] - y[4]	- y[5] + y[6] + y[7];	         //  <---------|* * *6* * * * * * * * O          |     4     |
//    b8 = y[0] - y[1] + y[2] - y[3] - y[4]	+ y[5] - y[6] + y[7];	         //            |           |         **          |           |
//                                                                           //            |           |        * *          |           |
//    c1 = z[0] + z[1] + z[2] + z[3] + z[4]	+ z[5] + z[6] + z[7];	         //            |           4-------*--*-----3----|-----------3
//    c2 = z[0] - z[1] - z[2] + z[3] + z[4]	- z[5] - z[6] + z[7];	         //            |          /       *   *          |          / 
//    c3 = z[0] + z[1] - z[2] - z[3] + z[4]	+ z[5] - z[6] - z[7];	         //            |         /       *    *          |         /  
//    c4 = z[0] + z[1] + z[2] + z[3] - z[4]	- z[5] - z[6] - z[7];	         //            9        /       3     *         10        /   
//    c5 = z[0] - z[1] + z[2] - z[3] + z[4]	- z[5] + z[6] - z[7];	         //            |       /       /      *          |       /    
//    c6 = z[0] - z[1] - z[2] + z[3] - z[4]	+ z[5] + z[6] - z[7];	         //            |      /       /       *          |      /     
//    c7 = z[0] + z[1] - z[2] - z[3] - z[4]	- z[5] + z[6] + z[7];	         //            |     4       /        1          |     2      
//    c8 = z[0] - z[1] + z[2] - z[3] - z[4]	+ z[5] - z[6] + z[7];            //            |    /       /         *          |    /       
//                                                                           //            |   /       /          *          |   /        
//    // setup initial guess				                 //            |  /       /           *          |  /         
//    answer[0] = answer[1] = answer[2] = 0.0;			         //            | /       /            *          | /          
//                                                                           //            |/       /             *          |/           
//    // apply Newton-Raphson to solve the problem			         //            1-------/--------1----------------2            
//    nite = 0;							         //                   /               |                       
//    do {								         //                  /                |                       
//      if ((++nite) > 10) {					         //             eta /                 |                       
//        fprintf (stderr,"nc_hexa_3d: no convergence after 10 iterations"); //                L                  V dzeta                 
//        return 0;
//      }
//      u = answer[0];
//      v = answer[1];
//      w = answer[2];
//      
//      // compute the residual
//      r[0] = a1 + u * a2 + v * a3 + w * a4 + u * v * a5 + u * w * a6 + v * w * a7 + u * v * w * a8 - 8.0 * xp;
//      r[1] = b1 + u * b2 + v * b3 + w * b4 + u * v * b5 + u * w * b6 + v * w * b7 + u * v * w * b8 - 8.0 * yp;
//      r[2] = c1 + u * c2 + v * c3 + w * c4 + u * v * c5 + u * w * c6 + v * w * c7 + u * v * w * c8 - 8.0 * zp;
//      
//      // check for convergence
//      if ((r[0]*r[0]+r[1]*r[1]+r[2]*r[2]) < 1.e-20) break;      // sqrt(1.e-20) = 1.e-10
//      
//      p[0] = a2 + v * a5 + w * a6 + v * w * a8;                            //         ***   JE TO ODVOZENO PRO TYTO BAZOVE FCE   ***
//      p[1] = a3 + u * a5 + w * a7 + u * w * a8;			         //    N[1] = 0.125 * (1.0 + ksi) * (1.0 + eta) * (1.0 + dzeta);
//      p[2] = a4 + u * a6 + v * a7 + u * v * a8;			         //    N[2] = 0.125 * (1.0 - ksi) * (1.0 + eta) * (1.0 + dzeta);
//      p[3] = b2 + v * b5 + w * b6 + v * w * b8;			         //    N[3] = 0.125 * (1.0 - ksi) * (1.0 - eta) * (1.0 + dzeta);
//      p[4] = b3 + u * b5 + w * b7 + u * w * b8;			         //    N[4] = 0.125 * (1.0 + ksi) * (1.0 - eta) * (1.0 + dzeta);
//      p[5] = b4 + u * b6 + v * b7 + u * v * b8;			         //    N[5] = 0.125 * (1.0 + ksi) * (1.0 + eta) * (1.0 - dzeta);
//      p[6] = c2 + v * c5 + w * c6 + v * w * c8;			         //    N[6] = 0.125 * (1.0 - ksi) * (1.0 + eta) * (1.0 - dzeta);
//      p[7] = c3 + u * c5 + w * c7 + u * w * c8;			         //    N[7] = 0.125 * (1.0 - ksi) * (1.0 - eta) * (1.0 - dzeta);
//      p[8] = c4 + u * c6 + v * c7 + u * v * c8;                            //    N[8] = 0.125 * (1.0 + ksi) * (1.0 - eta) * (1.0 - dzeta);
//      
//      // solve for corrections
//      solve_3 (p,r,delta);
//      
//      // update guess
//      answer[0] -= delta[0];
//      answer[1] -= delta[1];
//      answer[2] -= delta[2];
//      
//    } while (1);
//    
//    // test if inside
//    for (i=0;i<3;i++)
//      if ((answer[i] < -(1.0+zero)) || (answer[i] > (1.0+zero)))  return 0;
//    
//    return 1;
//  }

/// ****************************************************************************************
/// **************************   WWW.SOFTSURFER.COM  ALGORITHMS   **************************
/// ****************************************************************************************
/// Copyright 2000, softSurfer (www.softsurfer.com)
/// This code may be freely used and modified for any purpose providing that this copyright
/// notice is included with it. SoftSurfer makes no warranty for this code, and cannot be
/// held liable for any real or imagined damage resulting from its use. Users of this code
/// must verify correctness for their application.

// a Triangle is given by three points: Point V0, V1, V2
// a Polygon is given by:
//        int n = number of vertex points
//        Point* V[] = an array of points with V[n]=V[0], V[n+1]=V[1]

// Note: for efficiency low-level functions are declared to be inline.


//  // isLeft(): tests if a point is Left|On|Right of an infinite line.
//  //    Input:  three points P0, P1, and P2
//  //    Return: >0 for P2 left of the line through P0 and P1
//  //            =0 for P2 on the line
//  //            <0 for P2 right of the line
//  inline int
//  isLeft( Point P0, Point P1, Point P2 )
//  {
//      return ( (P1.x - P0.x) * (P2.y - P0.y)
//              - (P2.x - P0.x) * (P1.y - P0.y) );
//  }
//  //===================================================================
//  
//  // orientation2D_Triangle(): test the orientation of a triangle
//  //    Input:  three vertex points V0, V1, V2
//  //    Return: >0 for counterclockwise
//  //            =0 for none (degenerate)
//  //            <0 for clockwise
//  inline int
//  orientation2D_Triangle( Point V0, Point V1, Point V2 )
//  {
//      return isLeft(V0, V1, V2);
//  }
//  //===================================================================
//  
//  // area2D_Triangle(): compute the area of a triangle
//  //    Input:  three vertex points V0, V1, V2
//  //    Return: the (float) area of T
//  inline float
//  area2D_Triangle( Point V0, Point V1, Point V2 )
//  {
//      return (float)isLeft(V0, V1, V2) / 2.0;
//  }
//  //===================================================================
//  
//  // orientation2D_Polygon(): tests the orientation of a simple polygon
//  //    Input:  int n = the number of vertices in the polygon
//  //            Point* V = an array of n+1 vertices with V[n]=V[0]
//  //    Return: >0 for counterclockwise
//  //            =0 for none (degenerate)
//  //            <0 for clockwise
//  //    Note: this algorithm is faster than computing the signed area.
//  int
//  orientation2D_Polygon( int n, Point* V )
//  {
//      // first find rightmost lowest vertex of the polygon
//      int rmin = 0;
//      int xmin = V[0].x;
//      int ymin = V[0].y;
//  
//      for (int i=1; i<n; i++) {
//          if (V[i].y > ymin)
//              continue;
//          if (V[i].y == ymin) {    // just as low
//              if (V[i].x < xmin)   // and to left
//                  continue;
//          }
//          rmin = i;          // a new rightmost lowest vertex
//          xmin = V[i].x;
//          ymin = V[i].y;
//      }
//  
//      // test orientation at this rmin vertex
//      // ccw <=> the edge leaving is left of the entering edge
//      if (rmin == 0)
//          return isLeft( V[n-1], V[0], V[1] );
//      else
//          return isLeft( V[rmin-1], V[rmin], V[rmin+1] );
//  }
//  //===================================================================
//  
//  // area2D_Polygon(): computes the area of a 2D polygon
//  //    Input:  int n = the number of vertices in the polygon
//  //            Point* V = an array of n+2 vertices
//  //                       with V[n]=V[0] and V[n+1]=V[1]
//  //    Return: the (float) area of the polygon
//  float
//  area2D_Polygon( int n, Point* V )
//  {
//      float area = 0;
//      int   i, j, k;     // indices
//  
//      for (i=1, j=2, k=0; i<=n; i++, j++, k++) {
//          area += V[i].x * (V[j].y - V[k].y);
//      }
//      return area / 2.0;
//  }
//  //===================================================================

// area3D_Polygon(): computes the area of a 3D planar polygon
//    Input:  int n = the number of vertices in the polygon
//            Point* V = an array of n+2 vertices in a plane
//                       with V[n]=V[0] and V[n+1]=V[1]
//            Point N = unit normal vector of the polygon's plane
//    Return: the (float) area of the polygon
 //
double area3D_Polygon (int n, const PoinT **V, const VectoR *N)
{
  double area = 0;
  double an, ax, ay, az;  // abs value of normal and its coords
  int coord;              // coord to ignore: 1=x, 2=y, 3=z
  int i, j, k;            // loop indices
  
  // select largest abs coordinate to ignore for projection
  ax = (N->x>0 ? N->x : -N->x);     // abs x-coord
  ay = (N->y>0 ? N->y : -N->y);     // abs y-coord
  az = (N->z>0 ? N->z : -N->z);     // abs z-coord
  
  coord = 3;                     // ignore z-coord
  if (ax > ay) { if (ax > az) coord = 1; }   // ignore x-coord
  else         { if (ay > az) coord = 2; }   // ignore y-coord
  
  // compute area of the 2D projection
  for (i=1, j=2, k=0; i<=n; i++, j++, k++)
    switch (coord) {
    case 1:    area += (V[i]->y * (V[j]->z - V[k]->z));  continue;
    case 2:    area += (V[i]->x * (V[j]->z - V[k]->z));  continue;
    case 3:    area += (V[i]->x * (V[j]->y - V[k]->y));  continue;
    }
  
  // scale to get area before projection
  an = sqrt( ax*ax + ay*ay + az*az);  // length of normal vector
  switch (coord) {
  case 1:   area *= (an / (2*ax));   break;
  case 2:   area *= (an / (2*ay));   break;
  case 3:   area *= (an / (2*az));
  }
  
  return area;
}

//===================================================================
// http://www.softsurfer.com/Archive/algorithm_0105/algorithm_0105.htm
//    Line and Ray and Segment with defining points {Point P0, P1;}
//        (a Line is infinite, Rays and Segments start at P0)
//        (a Ray extends beyond P1, but a Segment ends at P1)
//    Plane with a point and a normal {Point V0; Vector n;}
//    Triangle with defining vertices {Point V0, V1, V2;}
//    Polyline and Polygon with n vertices {int n; Point *V;}
//        (a Polygon has V[n]=V[0])
//===================================================================
// intersect_RayTriangle(): intersect a ray with a 3D triangle
//    Input:  a ray R, and a triangle T
//    Output: *I = intersection point (when it exists)
//    Return: -1 = triangle is degenerate (a segment or point)
//             0 = disjoint (no intersect)
//             1 = intersect in unique point I1
//             2 = are in the same plane
int intersect_RayTriangle (double zero, const PoinT *P0, const PoinT *P1, const PoinT *V0, const PoinT *V1, const PoinT *V2, PoinT* I)
{
  VectoR u, v, n;             // triangle vectors
  VectoR dir, w0, w;          // ray vectors
  double r, a, b;             // params to calc ray-plane intersect
  
  // get triangle edge vectors and plane normal
  u.beP2P (V0, V1);
  v.beP2P (V0, V2);
  n.beVectProduct (&u, &v);       // cross product
  
  // triangle is degenerate
  if (u.give_length() < zero  ||  v.give_length() < zero  ||  n.give_length() < zero)
    return -1;
  
  dir.beP2P (P0, P1);            // ray direction vector
  w0.beP2P (V0, P0);
  
  a = - n.giveScalProduct(&w0);
  b =   n.giveScalProduct(&dir);
  // ray is parallel to triangle plane
  if (fabs(b) < zero) {
    if (fabs(a) < zero)  return 2;  // ray lies in triangle plane
    else                 return 0;  // ray disjoint from plane
  }
  
  // get intersect point of ray with triangle plane
  r = a / b;
  if (r < 0.0 || r > 1.0)      // ray goes away from triangle plane
    return 0;                  // => no intersect
  // for a ray, do not test if(r > 1.0)
  
  
  I->copy(P0)->add(dir.tms(r));  // intersect point of ray and plane
  
  // is I inside T?
  double uu, uv, vv, wu, wv, D;
  uu = u.giveScalProduct(&u);
  uv = u.giveScalProduct(&v);
  vv = v.giveScalProduct(&v);
  
  w.beP2P (V0, I);
  wu = w.giveScalProduct(&u);
  wv = w.giveScalProduct(&v);
  
  D = uv * uv - uu * vv;
  
  // get and test parametric coords
  double s, t;
  s = (uv * wv - vv * wu) / D;
  if (s < 0.0-zero ||  s      > 1.0+zero)  return 0;   // I is outside T
  
  t = (uv * wu - uu * wv) / D;
  if (t < 0.0-zero || (s + t) > 1.0+zero)  return 0;   // I is outside T
  
  return 1;                      // I is in T
}






//===================================================================


/**
   function computes derivatives of tri-linear base functions on hexahedron with respect of natural coordinate x
   
   @param n - array of derivatives of base functions
   @param y,z - natural coodinates
*/
void dx_bf_lin_hex_3d (double *n,double y,double z)
{
  n[0]=(1.0+y)*(1.0+z)/8.0;
  n[1]=(1.0+y)*(1.0+z)/(-8.0);
  n[2]=(1.0-y)*(1.0+z)/(-8.0);
  n[3]=(1.0-y)*(1.0+z)/8.0;
  n[4]=(1.0+y)*(1.0-z)/8.0;
  n[5]=(1.0+y)*(1.0-z)/(-8.0);
  n[6]=(1.0-y)*(1.0-z)/(-8.0);
  n[7]=(1.0-y)*(1.0-z)/8.0;
}



/**
   function computes derivatives of tri-linear base functions on hexahedron with respect of natural coordinate y
   
   @param n - array of derivatives of base functions
   @param x,z - natural coodinates
*/
void dy_bf_lin_hex_3d (double *n,double x,double z)
{
  n[0]=(1.0+x)*(1.0+z)/8.0;     
  n[1]=(1.0-x)*(1.0+z)/8.0;
  n[2]=(1.0-x)*(1.0+z)/(-8.0);
  n[3]=(1.0+x)*(1.0+z)/(-8.0);
  n[4]=(1.0+x)*(1.0-z)/8.0;
  n[5]=(1.0-x)*(1.0-z)/8.0;
  n[6]=(1.0-x)*(1.0-z)/(-8.0);
  n[7]=(1.0+x)*(1.0-z)/(-8.0);
}



/**
   function computes derivatives of tri-linear base functions on hexahedron with respect of natural coordinate z
   
   @param n - array of derivatives of base functions
   @param x,y - natural coodinates
*/
void dz_bf_lin_hex_3d (double *n,double x,double y)
{
  n[0]=(1.0+x)*(1.0+y)/8.0;
  n[1]=(1.0-x)*(1.0+y)/8.0;
  n[2]=(1.0-x)*(1.0-y)/8.0;
  n[3]=(1.0+x)*(1.0-y)/8.0;
  n[4]=(1.0+x)*(1.0+y)/(-8.0);
  n[5]=(1.0-x)*(1.0+y)/(-8.0);
  n[6]=(1.0-x)*(1.0-y)/(-8.0);
  n[7]=(1.0+x)*(1.0-y)/(-8.0);
}


void jac_3d (double &jac, Dvctr &x,Dvctr &y,Dvctr &z,
	     double xi,double eta,double zeta)
{
  long i;
  double d1,d2,d3,d4,d5,d6,d7,d8,d9,*nx,*ny,*nz;
  
  nx = new double [x.give_size()];
  ny = new double [x.give_size()];
  nz = new double [x.give_size()];
  
  switch (x.give_size()){
  //case 4:{
  //  dx_bf_lin_tet (nx);
  //  dy_bf_lin_tet (ny);
  //  dz_bf_lin_tet (nz);
  //  break;
  //}
  //case 6:{
  //  dx_bf_lin_wed_3d (nx,zeta);
  //  dy_bf_lin_wed_3d (ny,zeta);
  //  dz_bf_lin_wed_3d (nz,xi,eta);
  //  break;
  //}
  case 8:{
    dx_bf_lin_hex_3d (nx,eta,zeta);
    dy_bf_lin_hex_3d (ny,xi,zeta);
    dz_bf_lin_hex_3d (nz,xi,eta);
    break;
  }
    //case 10:{
  //  dx_bf_quad_tet (nx,xi,eta,zeta);
  //  dy_bf_quad_tet (ny,xi,eta,zeta);
  //  dz_bf_quad_tet (nz,xi,eta,zeta);
  //  break;
  //}
  //case 15:{
  //  dx_bf_quad_wed_3d (nx,xi,eta,zeta);
  //  dy_bf_quad_wed_3d (ny,xi,eta,zeta);
  //  dz_bf_quad_wed_3d (nz,xi,eta,zeta);
  //  break;
  //}
  //case 20:{
  //  dx_bf_quad_hex_3d (nx,xi,eta,zeta);
  //  dy_bf_quad_hex_3d (ny,xi,eta,zeta);
  //  dz_bf_quad_hex_3d (nz,xi,eta,zeta);
  //  break;
  //}
  default:{
    fprintf (stderr,"\n\n wrong number of nodes on 3D element in the function");
    fprintf (stderr,"\n jac_3d (%s, line %d).\n",__FILE__,__LINE__);
  }
  }
  
  d1=0.0;  d2=0.0;  d3=0.0;  d4=0.0;
  d5=0.0;  d6=0.0;  d7=0.0;  d8=0.0;  d9=0.0;
  for (i=0;i<x.give_size();i++) {
    d1+=nx[i]*x[i];
    d2+=ny[i]*x[i];
    d3+=nz[i]*x[i];
    d4+=nx[i]*y[i];
    d5+=ny[i]*y[i];
    d6+=nz[i]*y[i];
    d7+=nx[i]*z[i];
    d8+=ny[i]*z[i];
    d9+=nz[i]*z[i];
  }
  
  jac = d1*d5*d9 + d4*d8*d3 + d7*d2*d6 - d7*d5*d3 - d8*d6*d1 - d9*d4*d2;
  
  delete [] nx;  delete [] ny;  delete [] nz;
}


} // namespace gelibspace
