#ifndef GELIB_IO_H
#define GELIB_IO_H

#include "gelib.h"

#include <stdio.h>


/**
 * \file 
 * Input / output function. Strings, files, etc.
 */


#define _ARRAYS

namespace gelibspace {

class Dvctr;

//* ********************************************************************************************
//* *** *** *** ***                      GENERAL FUNCTIONS                       *** *** *** ***
//* ********************************************************************************************
/// non-word chars ==  ' ' '\t' '\n' '\0' '\r'
/// word chars     == !non-word chars
/// 
/// return true if character C is "normal" char                  --> C !=  ' '  '\t'  '\n'  '\0'  '\r'
bool isWordChar (char C);
/// return true if character C is "normal" char or "space" char  --> C !=             '\n'  '\0'  '\r'
bool isLineChar (char C);
/// return true if character C is "space" char                   --> C ==  ' '  '\t'
bool isSpaceChar (char C);

/// return length of line == sting compount of "line chars"
long giveLineLength (const char *src);

/// standard strdup  - Memory for the new string is obtained with malloc(3), and can be freed with free(3).
/// my       strdupl -                                            new                              delete 
char* strdupl (const char *src);
char* strdupl (const char *src1, const char *src2);
char* strdupl (const char *src1, const char *src2, const char *src3);
char* pathstrdupl (const char *src);


void errorr_expected (const char* file, int line, const char *msg, const char *expctd, const char *src);
void errorr_expected (const char* file, int line, const char *msg,         int expctd, const char *src);
void errorr_expected (const char* file, int line, const char *msg,        long expctd, const char *src);
void errorr_expected (const char* file, int line, const char *msg,      double expctd, const char *src);



//* ********************************************************************************************
//* *** *** *** ***                     ARGUMENTS PROCESSING                     *** *** *** ***
//* ********************************************************************************************
/// pass through arguments looking for option "s"  [former find_param]
long AP_find_option (int argc, char *argv[], const char *s);
/// print arguments to stream  [former print_command_line]
void AP_fprint_arguments (FILE *stream, int argc, char *argv[]);



//* ********************************************************************************************
//* *** *** *** ***                       FILE PROCESSING                        *** *** *** ***
//* ********************************************************************************************
/// move file descriptor to the start of the new noncommented word(digit) == skip nonword chars and comment
bool FP_skip_comment  (FILE *stream);
/// move file descriptor to the start of the new noncommented line, without checking of EOF
void FP_skip_line_commented  (FILE *stream);
/// move file descriptor to the start of the n-th new line, without checking of EOF
void FP_skip_line_fast_skip_commented  (FILE *stream, int n=1);
/// move file descriptor to the start of the n-th new line //[former read_line]
bool FP_skip_line  (FILE *stream, int n=1);
/// move file descriptor behind expected string
bool FP_skip_expected_string (FILE *src, const char *expctd, bool cs);
/// move file descriptor to the start of the new line after the one starting with string 
bool FP_skip_behind_line_starting_with (FILE *stream, const char *string, bool cs);

/// move file descriptor to the line starting with string, exactly behind the string
bool FP_skip_to_line_starting_with (FILE *stream, const char *string, bool cs);
/// move file descriptor to the line starting with string, exactly behind the string; NO white space can be before string
bool FP_skip_to_line_starting_With (FILE *stream, const char *string, bool cs);

/// move file descriptor to the line starting with character, directly to this char
bool FP_skip_to_line_starting_with (FILE *stream, char C);

///
void FP_copy_file (FILE *sour,FILE *dest);

/// copy from file to file behind the line [former copy_line]
bool FP_copy_behind_line (FILE *sour, FILE *dest, int n=1);
/// copy from file to file behind the line starting with string
bool FP_copy_behind_line_starting_with (FILE *src, FILE *dest, const char *expctd, bool cs);



/// scan/copy array of numbers      from src to dest, src pointer is shifted over the field
bool FP_scan_array (FILE *stream, int n, int    *dest);
bool FP_scan_array (FILE *stream, int n, long   *dest);
bool FP_scan_array (FILE *stream, int n, double *dest);

/// count number of words (see isWordChar) in file
long FP_number_of_words (FILE *stream);
/// count number of lines in file
long FP_number_of_lines (FILE *stream, bool rwd=false);


/// ... nonword == ' ', '\n', '\r' and '\t' characters
bool FP_skip_nonword (FILE *stream);

/// ... space == ' ' and '\t' characters
bool FP_skip_space (FILE *stream);

/// scan/copy line == string without '\n' from stream, skip lines commented by '#' at the first position
long FP_scan_line_skip_emptyORcommented (FILE *stream, char *dest);

/// scan/copy line == string without '\n' from stream
long FP_scan_line (FILE *stream, char *dest);

/// scan/copy line == string without '\n' from stream; dest is NULL and allocated here
long FP_scan_line_alloc (FILE *stream, char *&dest);

/// ... word; return value is length of the word
int FP_scan_word (FILE *stream, char *dest);

/// ... word and compare with expected one
bool FP_scan_expected_word (FILE *src, const char *expctd, bool cs=true);

/// ... word and compare with expected one, exit if false
void FP_scan_expected_word_FL (const char* file, int line, FILE *src, const char *expctd, const char *msg, bool cs);
#define FP_scan_expected_word_exit(_1,_2,_3,_4) FP_scan_expected_word_FL (__FILE__, __LINE__, _1,_2,_3,_4)
/// ... line and compare with expected one, exit if false
void FP_scan_expected_line_FL (const char* file, int line, FILE *src, const char *expctd, const char *msg, bool cs);
#define FP_scan_expected_line_exit(_1,_2,_3,_4) FP_scan_expected_line_FL (__FILE__, __LINE__, _1,_2,_3,_4)


/// ... number and compare with expected one
bool FP_scan_expected_number (FILE *src, long expctd);

/// ... number and compare with expected one, exit if false
template <class ArgType>
void FP_scan_expected_number_FL (const char* file, int line, FILE *src, ArgType expctd, const char *msg)
{
  if (!FP_scan_expected_number (src, expctd))
    errorr_expected (file, line, msg, expctd, ""); // misto "" by tam mel byt retezec s prectenym cislem, casem to dodelej, asi budes muset predelat errorr_expected
}
#define FP_scan_expected_number_exit(_1,_2,_3) FP_scan_expected_number_FL (__FILE__, __LINE__, _1,_2,_3)




///// scan number and compare with expected one
//void FP_scan_expected_long number (FILE *stream, ArgType expctd)
//{
//  ArgType aux;
//  SP_scan_number (src, aux);
//  if (aux != expctd)  return false;
//  return true;
//  
//fscanf (in, "%ld", &auxl);  if ( auxl != -1)  _errorr3 ("Invalid structure of the given UNV file: actual != expected -- %ld != %ld", -1, auxl);
//
//
//}



//* ********************************************************************************************
//* *** *** *** ***                      STRING PROCESSING                       *** *** *** ***
//* ********************************************************************************************

/// cut of comment == end of line starting by '#'
bool SP_cutcomment (char *src);

/// replace src char to dest char
bool SP_replace_char (char *str, char s, char d);

//* *******************************************
//* ***   ***   ***   SKIPING   ***   ***   ***
//* *******************************************
/// skiping == shift src pointer at the end of ...

/// ... space == ' ' and '\t' characters
bool SP_skip_space (const char *&src);
/// ... word and space before
bool SP_skip_word (const char *&src, int n=1);
/// ... word compound of digits and space before
bool SP_skip_int (const char *&src);


//* ********************************************
//* ***   ***   ***   SCANNING   ***   ***   ***
//* ********************************************
/// unscan word == shift pointer back over word
void SP_unscan_word (const char *&src, const char *w);

/// scanning == scan ... to variable dest and shift src pointer at the end of ...

/// ... one char
bool SP_scan_character (const char *&src, char &dest);
/// ... word; return value is length of the word
long SP_scan_word      (const char *&src, char *dest);
/// ... line; return value is length of the line
long SP_scan_line      (const char *&src, char *dest);
/// ... line; return value is length of the line; dest is NULL and allocated here
long SP_scan_line_alloc (const char *&src, char *&dest);

/// ... word and compare with expected one
bool SP_scan_expected_word (const char *&src, const char *expctd, bool cs=true);



/// ... word and compare with expected one/two/three, exit if false
void SP_scan_expected_word_FL (const char* file, int line, const char *&src, const char *expctd,                                            const char *msg, bool cs);
void SP_scan_expected_word_FL (const char* file, int line, const char *&src, const char *expctd1, const char *expctd2,                      const char *msg, bool cs);
void SP_scan_expected_word_FL (const char* file, int line, const char *&src, const char *expctd1, const char *expctd2, const char *expctd3, const char *msg, bool cs);
//
#define SP_scan_expected_word_exit( _1,_2,_3,_4)       SP_scan_expected_word_FL (__FILE__, __LINE__, _1,_2,_3,_4)
#define SP_scan_expected_word2_exit(_1,_2,_3,_4,_5)    SP_scan_expected_word_FL (__FILE__, __LINE__, _1,_2,_3,_4,_5)
#define SP_scan_expected_word3_exit(_1,_2,_3,_4,_5,_6) SP_scan_expected_word_FL (__FILE__, __LINE__, _1,_2,_3,_4,_5,_6)



/// ... number of type int/long/double
bool SP_scan_number (const char *&src, int    &dest);
bool SP_scan_number (const char *&src, long   &dest);
bool SP_scan_number (const char *&src, double &dest);

/// ... number and compare with expected one
template <class ArgType>
bool SP_scan_expected_number (const char *&src, ArgType expctd)
{
  ArgType aux;
  return (SP_scan_number (src, aux)  &&  aux == expctd);
}

/// ... number and compare with expected one, exit if false
template <class ArgType>
void SP_scan_expected_number_FL (const char* file, int line, const char *&src, ArgType expctd, const char *msg)
{
  if (!SP_scan_expected_number (src, expctd))
    errorr_expected (file, line, msg, expctd, src);
}
#define SP_scan_expected_number_exit(_1,_2,_3) SP_scan_expected_number_FL (__FILE__, __LINE__, _1,_2,_3)


/// ... array of numbers
template <class ArgType> bool SP_scan_array (const char *&src, int  n, ArgType *a) {  while (--n)  SP_scan_number (src, (a++)[0]);  return SP_scan_number (src, a[0]); }
template <class ArgType> bool SP_scan_array (const char *&src, long n, ArgType *a) {  while (--n)  SP_scan_number (src, (a++)[0]);  return SP_scan_number (src, a[0]); }

template <class ArgType>
bool SP_scan_array_alloc (const char *&src, long &n, ArgType *&a)
{
  SP_scan_number (src, n);
  a = new ArgType[n];
  return SP_scan_array (src, n, a);
}

/// 
bool SP_scan_array_L (const char *&src, int L, double *a);
///

#ifdef _ARRAYS
bool SP_scan_Dvctr_exit (const char *&src, int n, Dvctr *v);
#endif


/// bool SP_scan_Dvctr      (const char *&src,        Dvctr *dest);
template <class ArgType>
bool SP_scan_Xvctr (const char *&src, ArgType *dest)
{
  long n;
  SP_scan_number (src, n);
  dest->resize_ignore_vals(n);
  return SP_scan_array (src, n, dest->give_ptr2val());
}



//* ********************************************
//* ***   ***   ***   PRINTING   ***   ***   ***
//* ********************************************
/// printing == print ... to stream and shift src pointer at the end of ...

/// ... space == ' ' and '\t' characters
bool SP_print_space (const char *&src, FILE *stream);
/// ... word
bool SP_print_word (const char *&src, FILE *stream);


//* ********************************************************************************************
//* *** *** *** ***                      OBSOLETE FUNCTIONS                      *** *** *** ***
//* ********************************************************************************************
long find_double (long n, long* p);

void sprint_param (char *s, char *argv[], int n);

void read_star_end (FILE *stream,int n=1);
void read_star (FILE *stream);
/// read line of file until \n, no save, dava bacha na EOF

void copy_star_end (FILE *sour,FILE *dest,int n=1);
void copy_star (FILE *sour,FILE *dest);

void copy_to_end_line (FILE *sour,FILE *dest);


int  extract_long   (const char *s1, const char *s2, long &a);
int  extract_double (const char *s1, const char *s2, double &a);

void extract_double_array (const char *s1, const char *s2, long n, double *a);

void give_line (FILE *stream,char *&s);
void ad_line (FILE *stream,char *s);

} // namespace gelibspace

#endif // GELIB_IO_H
