#include "tixy2.h"

#include "librw.h"


namespace gelibspace {

/// ********************************************************************************************
/// *** *** *** ***                         TINYXML FCE                          *** *** *** ***
/// ********************************************************************************************


//XMLDocument* openTXdcPNSS (const char *path, const char *name, const char *suff, const char *suff2)
//{
//  char file[255];
//  if (path)  sprintf (file, "%s%s%s%s", path, name, suff, suff2);
//  else       sprintf (file,   "%s%s%s",       name, suff, suff2);
//  
//  return new XMLDocument(file);
//}

//XMLDocument* openTXdcPN (const char *path, const char *name)
//{
//  char file[255];
//  if (path)  sprintf (file, "%s%s", path, name);
//  else       sprintf (file,   "%s",       name);
//  
//  return new XMLDocument(file);
//}



/// ********************************************************************************************
/// *** *** *** ***                      STREAM PROCESSING                       *** *** *** ***
/// ********************************************************************************************
bool ST_scan_number (Stream *src, int &dest)
{
  if (src->isFile  ())  return ( fscanf (src->file(), "%d", &dest) == 1 );
  if (src->isString())  return SP_scan_number (*src->pstring(), dest);
  errol;  return false;
}
bool ST_scan_number (Stream *src, long &dest)
{
  if (src->isFile  ())  return ( fscanf (src->file(), "%ld", &dest) == 1 );
  if (src->isString())  return SP_scan_number (*src->pstring(), dest);
  errol;  return false;
}
bool ST_scan_number (Stream *src, double &dest)
{
  if (src->isFile  ())  return ( fscanf (src->file(), "%lf", &dest) == 1 );
  if (src->isString())  return SP_scan_number (*src->pstring(), dest);
  errol;  return false;
}

bool ST_scan_array (Stream *src, int n, int    *dest)
{
  if (src->isFile  ())  return FP_scan_array (src->file(), n, dest);
  if (src->isString())  return SP_scan_array (*src->pstring(), n, dest);
  errol;  return false;
}
bool ST_scan_array (Stream *src, int n, long   *dest)
{
  if (src->isFile  ())  return FP_scan_array (src->file(), n, dest);
  if (src->isString())  return SP_scan_array (*src->pstring(), n, dest);
  errol;  return false;
}
bool ST_scan_array (Stream *src, int n, double *dest)
{
  if (src->isFile  ())  return FP_scan_array (src->file(), n, dest);
  if (src->isString())  return SP_scan_array (*src->pstring(), n, dest);
  errol;  return false;
}

/// ... number and compare with expected one
bool ST_scan_expected_number (Stream *src, long expctd)
{
  long aux;
  return (ST_scan_number(src, aux)  &&  aux == expctd);
}



/// ... line; return value is length of the line
long GP_scan_line (Stream * stream, char *dest)
{
  if (stream->isFile())
    return FP_scan_line (stream->file(), dest);
  else {
    const char *a = stream->tixel()->GetText();
    return SP_scan_line (a, dest);
  }
}

/// ... line; return value is length of the line; dest is NULL and allocated here
long GP_scan_line_alloc (Stream * stream, char *&dest)
{
  if (stream->isFile())
    return FP_scan_line_alloc (stream->file(), dest);
  else {
    const char *a = stream->tixel()->GetText();
    return SP_scan_line_alloc (a, dest);
  }
}
long GP_scan_line_alloc_skip_commented (Stream * stream, char *&dest)
{
  if (stream->isFile())  FP_skip_line_commented (stream->file());
  
  return GP_scan_line_alloc (stream, dest);
}

/// ... word; return value is length of the word
long GP_scan_word (Stream * stream, char *dest)
{
  if (stream->isFile())
    return FP_scan_word (stream->file(), dest);
  else {
    const char *a = stream->tixel()->GetText();
    return SP_scan_word (a, dest);
  }
}

//* ********************************************************************************************
//* *** *** *** ***                       docasne XML FCES                       *** *** *** ***
//* ********************************************************************************************
const XMLElement * XP_give_unique_expected_elem (const XMLNode *xelem, const char * name, bool uniq)
{
  const XMLElement * answer = xelem->FirstChild()->ToElement();
  if (!answer)                               _errorr  ("no child");
  if (uniq && answer->NextSibling() != NULL) _errorr  ("is not unique");
  if ( strcmp(answer->Value(), name) != 0 )  _errorr2 ("name of xelement is not \"%s\"", name);

  return answer;
}
void XP_check_expected_attribute (const XMLNode *xelem, const char *name, const char *value)
{
  if ( strcmp(xelem->ToElement()->Attribute(name), value) != 0 )  _errorr2 ("no attribute \"%s\"", value);
}

//int XP_give_attribute_int (const XMLElement *xelem, const char *name)
//{
//  int answer;
//  if (xelem->QueryIntAttribute(name, &answer) != TIXML_SUCCESS)  _errorr ("no attribute \"%s\"", name);
//  return answer;
//}
const char * XP_giveDAtext (const XMLNode *xelem, int n, bool last, const char *format, const char *type, const char *name, int *noc)
{
  const XMLNode *da = NULL;
  if      (n == 0)  da = xelem;
  else if (n == 1)  da = xelem->FirstChild();
  else if (n == 2)  da = xelem->FirstChild()->NextSibling();
  else if (n == 3)  da = xelem->FirstChild()->NextSibling()->NextSibling();
  else _errorr ("error");
  
  if (!da)                                      _errorr ("no data array");
  if (last && da->NextSibling() != NULL)        _errorr ("is not last");
  if ( strcmp(da->Value(), "DataArray") != 0 )  _errorr ("name of xelement is not DataArray");
  
  ;          XP_check_expected_attribute (da, "format", format);
  ;          XP_check_expected_attribute (da, "type", type);
  if (name)  XP_check_expected_attribute (da, "Name", name);
  if (noc) {
    const char *no = da->ToElement()->Attribute("NumberOfComponents");
    int n;
    if (no)  n = atol(no);
    else     n = 1;
    if (*noc && *noc != n)  _errorr3 ("in DataArray, NumberOfComponents is not \"%s\" but %s", *noc, n);
    else *noc = n;
  }
  
  const char * text = da->ToElement()->GetText();
  if (text == NULL)
    _errorr("no text in DataArray");
  
  return text;
}


} // namespace gelibspace
