#ifndef STRUT_XALM_H
#define STRUT_XALM_H

#include "xalm.h"



/**
 * \file
 * Interface to library XALM.
 */

namespace xalm {


//   |           | F = 1.0 
//   |           | 
//   |           V
// 3 +           +<|
//   |         /
//   |       /
//   |     /
//   |   /
//   | /
//   +-----------+----
//  0.0         4.0 
//
//
//   
//
//

  ////
  //// vzperadlo - ukazkovy priklad pouziti knihovny XALM
  ////


class XALM_interface : public XALM
{
 private:
  // promenne, ktere jsou potrebne jen pro tento priklad, tj. vzperadlo
  //int lm[4] = {1,2,3,4};
  double xze[4];
  double E, A;
  bool tangent_stiff_mtrx;
  
  double k;
  
  double *tisk1;
  double *tisk2;
  
  
 public:
  ///  CONSTRUCTOR
  XALM_interface (void) {}
  ///  DESTRUCTOR
  virtual ~XALM_interface (void) { delete [] tisk1; delete [] tisk2; }
  
  ///
  void mtlb (void);
  
  
 protected:
  /// Funkce nemá parametry a nic nevrací. Má za úkol naplnit chráněné atributy v kterých se uchovávají základní, ve třídě XALM často používané, charakteristiky řešené úlohy.
  /// Atributy typu ukazatel na pole musí být alokovány uvnitř této funkce. Dealokovány jsou v destruktoru třídy XALM.
  virtual void initialize (void);
  
  /// Ve funkci je možné po každém provedeném iteračním kroku vykonat požadované úkony.
  /// Např. vypsat aktuání hodnoty, zastavit iteraci, pokud je dosaženo požadovaných výsledků, atd.
  virtual void update_step (void);
  
  ///
  virtual void update_stiffness_matrix (const Dvctr *X);
  ///
  virtual void update_internal_forces (Dvctr *internalForces, const Dvctr *X);
  ///
  virtual void lineq_solve (Dvctr *X, const Dvctr *R);
  
  
 private:
  ///
  void initialize_local (void);
  
};

} // end namespace xalm

#endif // STRUT_XALM_H
